#include <avr/io.h>
#include <avr/interrupt.h>
#include "uart.h"
#include "m8help.h"

#ifndef _ECL_H_
#define _ECL_H_

//Informationen zum Eingang - muss Interrupt 0 sein sein.
#define ECL_PINREG PIND
#define ECL_PIN PD2
#define ECL_DDR DDRC
#define ECL_SIG_INVERTED 1

//Externe Gerte, die verwendet werden
#define ECL_USE_ECA86 1
#define ECL_USE_ECA60 1

//ECL-Paket zu Debugging-Zwecken ausgeben
#define ECL_DUMP_PACKET 1
#define ECL_DUMP_CHR(c) uart_putc(c)
#define ECL_DUMP_HEX(c) uart_puthex(c)

//Erweiterte Debugging-Funktionalitt (nicht wirklich optimiert)
#define ECL_DEBUG 0

#if ECL_DEBUG == 1
	#define ECL_DEBUG_C(c) uart_putc(c)
	#define ECL_DEBUG_S(s) uart_puts(s)
	#define ECL_DEBUG_I(i) uart_putnum(i)
	#define ECL_DEBUG_H(h) uart_puthex(h)
#else
	#define ECL_DEBUG_C(ignored)
	#define ECL_DEBUG_S(ignored)
	#define ECL_DEBUG_I(ignored)
	#define ECL_DEBUG_H(ignored)
#endif



//Prescaler fr den Timer, sollte nur im Zusammenspiel mit ECL_BIT_USEC und des CPU-Takts verndert werden
#define ECL_TMR_PRESCALER 64

//Bitlnge in s
#define ECL_BIT_USEC 409
#define ECL_BITLEN ((F_CPU/ECL_TMR_PRESCALER*ECL_BIT_USEC) / 1000000 + 1)

//Prfen, ob die Bitlnge korrekt ist
#if ECL_BITLEN > 255
	#warning ECL_BITLEN is too big, change Prescaler!
#endif
#if ECL_BITLEN < 20
	#warning ECL_BITLEN might be to small, please try another Prescaler!
#endif


/*** Ab hier bitte nichts mehr verndern ***/

#define ECL_PACKET_SIZE 11

//States/Return-Values beim Dekodieren
#define ECL_S_NOT_HANDLED    0x00
#define ECL_S_NOT_READY      0x01
#define ECL_S_NOT_COMPLETE   0x02
#define ECL_S_CHECKSUM_ERROR 0x03
#define ECL_S_TMP_OUTDOOR    0x04
#define ECL_S_TMP_INDOOR     0x05
#define ECL_S_TMP_ECA86_1    0x06
#define ECL_S_TMP_ECA86_2    0x07
#define ECL_S_SETPOINTMODE   0x08
#define ECL_S_TIME           0x09
#define ECL_S_DAYPROGRAM     0x0A
#define ECL_S_RELAIS         0x0B

#define ECL_S_NORMAL         0x00		
#define ECL_S_REQUEST        0x40
#define ECL_S_SET            0x80

//Buffer-Positionen
#define ECL_B_ADDR 0
#define ECL_B_PTYPE 1

//Quell-Adressen
#define ECL_A_SRC_ECA60  0xA0
#define ECL_A_SRC_ECL300 0xF0
#define ECL_A_SRC_ECA86  0xE0

//Ziel-Adressen
#define ECL_A_DST_ECA60  0x0A
#define ECL_A_DST_ECL300 0x0F
#define ECL_A_DST_ECA86  0x0E
#define ECL_A_DST_BCAST  0x00

//Kombinierte Adressen
#define ECL_A_ECA60_ECL300  (ECL_A_SRC_ECA60  | ECL_A_DST_ECL300)
#define ECL_A_ECL300_ECA60  (ECL_A_SRC_ECL300 | ECL_A_DST_ECA60)
#define ECL_A_ECL300_BCAST  (ECL_A_SRC_ECL300 | ECL_A_DST_BCAST)
#define ECL_A_ECA86_ECL300  (ECL_A_SRC_ECA86  | ECL_A_DST_ECL300)
#define ECL_A_ECL300_ECA86  (ECL_A_SRC_ECL300 | ECL_A_DST_ECA86)
//sieht komisch aus, passiert aber manchmal...
#define ECL_A_ECL300_ECL300 (ECL_A_SRC_ECL300 | ECL_A_DST_ECL300)

//Pakettypen
#define ECL_P_TMP_INDOOR  0x04
#define ECL_P_TMP_OUTDOOR 0x01
#define ECL_P_TMP_ECA86   0x60
#define ECL_P_SET_RELAIS  0x62
#define ECL_P_SET_MODE    0x05
#define ECL_P_GET_MODE    0x06
#define ECL_P_TIME        0x02
#define ECL_P_SET_TIME    0x11
#define ECL_P_REQ_PROG    0x09
#define ECL_P_PROG        0x09

#define ECL_P_TMP_ECA86_0 0x23
#define ECL_P_TMP_ECA86_1 0x45

//Arten der An- bzw. Abwesenheit
#define ECL_DIFF_NONE  1
#define ECL_DIFF_RELAX 2
#define ECL_DIFF_AWAY  3

typedef struct {
	uint8_t	second;
	uint8_t	minute;
	uint8_t	hour;
	uint8_t	day;
	uint8_t	wday;
	uint8_t	month;
	uint16_t year;
} time_t;

typedef struct {
	int16_t tmp_outdoor;
#if ECL_USE_ECA60 == 1
	int16_t tmp_indoor;
	int8_t  tmp_diff;
	uint8_t prog_day:3;
	uint8_t prog[6];
#endif
	uint8_t tmp_setpoint;
#if ECL_USE_ECA86 == 1
	int16_t tmp_ECA86[4];
	uint8_t relais;
#endif
	uint8_t mode_water:4;
	uint8_t mode_heating:4;
	uint8_t mode:3;
	time_t time;
} ecl_data_t;

extern ecl_data_t ecl_data;
extern volatile uint8_t ecl_bytepos;
extern volatile uint8_t ecl_pos;
extern volatile uint8_t ecl_packet[ECL_PACKET_SIZE];

void ecl_init();
uint8_t ecl_decode();
static inline void ecl_int(void) __attribute__((always_inline));
static inline void ecl_tmr_ovf(void) __attribute__((always_inline));

//Sowas hat in der Headerdatei eigentlich nichts zu suchen, 
//aber anscheinend sind nur so eine "echte" inline-Funktionen mglich.

//Interrupt beim Flankenwechsel an INT0, startet die Erfassung des Pakets
void ecl_int() {
	INT0_IRQ_DISABLE(); //zum samplen deaktivieren
	
	//Erstes Bit (Startbit) berspringen und auf die Mitte des Bits samplen
	TCNT0 = 256 - (ECL_BITLEN + ECL_BITLEN/2);
	TIMER0_IRQ_CLEAR();
	TIMER0_IRQ_ENABLE();
	ecl_pos = 0;
}

//Sampeln des Pakets
void ecl_tmr_ovf() {
	TCNT0 = 256 - ECL_BITLEN; 

	if(ecl_pos < 15) {

		//Byte vor dem beschreiben leeren
		if(ecl_pos == 0 || ecl_pos == 8) { ecl_packet[ecl_bytepos] = 0; } 

		//Bit einlesen und im Puffer speichern
#if ECL_SIG_INVERTED == 1
		if((ECL_PINREG & (1<<ECL_PIN)) == 0) {
#else
		if((ECL_PINREG & (1<<ECL_PIN)) == (1<<ECL_PIN)) {
#endif
		
			uint8_t buff_bit = 1<<(ecl_pos & 0x07);

			ecl_packet[ecl_bytepos] |= buff_bit;
		}
		if(ecl_pos == 7) {
			ecl_bytepos++;
			if(ecl_bytepos == ECL_PACKET_SIZE) { ecl_bytepos = 0; }
		}
	} else if(ecl_pos == 15) {
		//Letztes Bit nicht mehr einlesen
		TCNT0 = 0; //Holdoff-Zeit, damit nicht sofort wieder gelesen wird
		ecl_bytepos++;
		if(ecl_bytepos == ECL_PACKET_SIZE) { ecl_bytepos = 0; }
	} else if(ecl_pos == 16) {
		//Timer deaktivieren, INT0 wieder freigeben
		TIMER0_IRQ_DISABLE();
		INT0_IRQ_CLEAR();
		INT0_IRQ_ENABLE();
	}
	ecl_pos++;
}

#endif
